/* This code is taken from write.c, in the util-linux 1.10 distribution
 * (originally also BSD).  write does much better than talk for people 
 * logged on >1 time
 *
 * The two programs hacked together by Daniel Barlow 
 * <Daniel.Barlow@sjc.ox.ac.uk>, Fri Feb 10 02:35:51 GMT 1995    
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/stat.h>		/* many of these are probably */
#include <sys/file.h>		/* unnecessary.  I just copied */
#include <sys/time.h>		/* the lot from write.c */
#include <utmp.h>
#include <ctype.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <protocols/talkd.h>
#include <syslog.h>
#include <errno.h>
 
#ifndef UT_LINESIZE		/* apparently undefined in linux */
#define UT_LINESIZE (sizeof(u.ut_line))
#endif

/*
 * search_utmp - search utmp for the "best" terminal to write to
 *
 * Ignores terminals with messages disabled, and of the rest, returns
 * the one with the most recent access time.  Returns as value the number
 * of the user's terminals with messages enabled, or -1 if the user is
 * not logged in at all.
 * 
 * The code for talking to oneself was deleted as I'm not sure that the
 * 'mytty' equivalent is available
 * 
 * Special case for writing to yourself - ignore the terminal you're
 * writing from, unless that's the only terminal with messages enabled.  
 */

extern int debug;

search_utmp(user, tty)		/* tty is a return thingy not an argument */
	char *user, *tty;
{
	struct utmp u;
	time_t bestatime, atime;
	int ufd, nloggedttys, nttys, msgsok;

	char atty[UT_LINESIZE + 1];

	if ((ufd = open(_PATH_UTMP, O_RDONLY)) < 0) {
		perror(_PATH_UTMP);
		return (FAILED);
	}

	nloggedttys = nttys = 0;
	bestatime = 0;
	while (read(ufd, (char *) &u, sizeof(u)) == sizeof(u))
		if (strncmp(user, u.ut_name, sizeof(u.ut_name)) == 0) {
			++nloggedttys;
			(void)strncpy(atty, u.ut_line, UT_LINESIZE);
			atty[UT_LINESIZE] = '\0';

			if (term_chk(atty, &msgsok, &atime, 0) || !(msgsok))
				continue;	/* bad term? skip */

#ifdef __linux__
                        if (u.ut_type != USER_PROCESS)
			        continue;       /* it's not a valid entry */
#endif
			++nttys;
			if (atime > bestatime) {
				bestatime = atime;
				(void)strcpy(tty, atty);
			}
		}

	(void)close(ufd);
	if (nloggedttys == 0) {
	        if(debug) print_err("talk: %s is not logged in", user);
		return NOT_HERE; 
	}
	if (nttys == 0) {
		if(debug) print_err("talk: %s has messages disabled",user);
		return PERMISSION_DENIED;
	}
	return  SUCCESS ;
}

/*
 * term_chk - check that a terminal exists, and get the message bit
 *     and the access time
 */
term_chk(tty, msgsokP, atimeP, showerror)
	char *tty;
	int *msgsokP, showerror;
	time_t *atimeP;
{
	struct stat s;
	char path[MAXPATHLEN];

	(void)sprintf(path, "/dev/%s", tty);
	if (stat(path, &s) < 0) {
		if (debug)  
		  print_err("term_chk: %s", strerror(errno));
		return(1);
	}
	*msgsokP = (s.st_mode & (S_IWRITE >> 3)) != 0;	/* group write bit */
	*atimeP = s.st_atime;
	return(0);
}

/* simple minded error printing via syslog */

print_err(msg,user)
char *msg, *user;
{
	char *buf=malloc(strlen(msg)+10);
	sprintf(buf,msg,user);
	syslog(LOG_DEBUG, buf);
	free(buf);
	return 0;
}

 	
